//
// Simple templated Lua wrapper.
// Based on code from http://www.ilovelua.narod.ru
//

#ifndef	__LUA_TEMPLATES__
#define	__LUA_TEMPLATES__

#ifdef	__cplusplus
extern "C" {
#endif

#include	"lua.h"
#include	"lauxlib.h"
#include	"lualib.h"

#ifdef	__cplusplus
};
#endif

#include	<string>
#include	<map>

template<typename T> bool fromLua ( lua_State * lua, int index, T& ret );
template<typename T> void toLua   ( lua_State * lua, T& arg );

template<> bool fromLua ( lua_State * L, int index, bool& ret )
{
	if( !lua_isboolean ( L, index ) )
		return false;
		
	ret = lua_toboolean(L, index) != 0;
    return true;
}

template<> bool fromLua (lua_State * lua, int index, float& ret)
{
	if( !lua_isnumber ( lua, index ) )
		return false;
		
	ret = (float) lua_tonumber ( lua, index );
	return true;
}

template<> bool fromLua(lua_State* L, int index, double& ret)
{
	if(!lua_isnumber(L, index))
		return false;
		
	ret = lua_tonumber(L, index);
	return true;
}

template<> bool fromLua(lua_State* L, int index, int& ret)
{
	if( !lua_isnumber(L, index))
		return false;
 
	ret = (int)lua_tonumber(L, index);
	return true;
}

template<> bool fromLua(lua_State* L, int index, std::string& ret)
{
	if( !lua_isstring(L, index))
		return false;
		
	ret = lua_tostring(L, index);
	return true;
}

template<typename Key, typename Value>
bool fromLua ( lua_State * lua, int index, std::map<Key, Value>& ret)
{
	if(!lua_istable ( lua, index))
		return false;

	lua_pushvalue ( lua, index ); 		// stack: map
	lua_pushnil   ( lua ); 				// stack: map nil
	
	while ( lua_next ( lua, -2 ) ) 		// stack: map key value
	{
		Key 	key;
		Value	value;
		
		fromLua ( lua, -2, key );
		fromLua ( lua, -1, value);
		
		ret [key] = value;
		
		lua_pop ( lua, 1 ); 			// stack: map key
	}
	
	lua_pop ( lua, 1); 					// stack:

  return true;
}

template<typename T>
bool fromLuaTable ( lua_State * lua, int index, const char * key, T& ret)
{
	lua_getfield ( lua, index? index : LUA_GLOBALSINDEX, key ); // stack: table value

	bool res = fromLua ( lua, -1, ret);

	lua_pop ( lua, 1 ); // stack: table

	return res;
}

template<> void toLua ( lua_State * lua, const bool& arg )
{
	lua_pushboolean ( lua, arg ? 1 : 0 );
}

template<> void toLua ( lua_State * lua, bool& arg )
{
	lua_pushboolean ( lua, arg ? 1 : 0 );
}

template<> void toLua ( lua_State * lua, const float& arg )
{
	lua_pushnumber ( lua, arg );
}

template<> void toLua ( lua_State * lua, float& arg )
{
	lua_pushnumber ( lua, arg );
}

template<> void toLua ( lua_State * lua, const double& arg )
{
	lua_pushnumber ( lua, arg );
}

template<> void toLua ( lua_State * lua, double& arg )
{
	lua_pushnumber ( lua, arg );
}

template<> void toLua ( lua_State * lua, const int& arg )
{
	lua_pushnumber ( lua, arg );
}

template<> void toLua ( lua_State * lua, int& arg )
{
	lua_pushnumber ( lua, arg );
}

template<> void toLua ( lua_State * lua, const std::string& arg )
{
	lua_pushstring ( lua, arg.c_str () );
}

template<> void toLua ( lua_State * lua, std::string& arg )
{
	lua_pushstring ( lua, arg.c_str () );
}

/////////////// templates to call Lua functions passsing arguments /////////////////////

template <typename R> 
static inline bool callLua0 ( lua_State * lua, const char * funcName, R& res )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );	// push global function f on stack
	lua_pcall    ( lua, 0, 1, 0 );                   	// call function taking no argsuments and getting one return value
	fromLua      ( lua, -1, res );
	lua_pop      ( lua, 1 );
	
	return res;
}

template <typename R> 
static inline void	callLua0 ( lua_State * lua, const char * funcName )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );	// push global function f on stack
	lua_pcall    ( lua, 0, 1, 0 );                   	// call function taking no argsuments and getting one return value
}

template <typename R, typename A> 
static inline bool	callLua1 ( lua_State * lua, const char * funcName, R& res, A arg )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );   // push global function f on stack
	toLua        ( lua, arg );                   		// push first argument on stack
	lua_pcall    ( lua, 1, 1, 0 );                   	// call function taking 1 argsuments and getting one return value
	fromLua      ( lua, -1, res );
	lua_pop      ( lua, 1 );
	
	return true;
}

template <typename A> 
static inline void	callLua1 ( lua_State * lua, const char * funcName, A arg )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );   // push global function f on stack
	toLua        ( lua, arg );                   		// push first argument on stack
	lua_pcall    ( lua, 1, 1, 0 );                   	// call function taking 1 argsuments and getting one return value
}

template <typename R, typename A1, typename A2> 
static inline bool	callLua2 ( lua_State * lua, const char * funcName, R& res, A1 arg1, A2 arg2 )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );   // push global function f on stack
	toLua        ( lua, arg1 );                   		// push first argument on stack
	toLua        ( lua, arg2 );                   		// push 2nd argument on stack
	lua_pcall    ( lua, 2, 1, 0 );                   	// call function taking 2 argsuments and getting one return value
	fromLua      ( lua, -1, res );
	lua_pop      ( lua, 1 );
	
	return true;
}

template <typename A1, typename A2> 
static inline void	callLua2 ( lua_State * lua, const char * funcName, A1 arg1, A2 arg2 )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );   // push global function f on stack
	toLua        ( lua, arg1 );                   		// push first argument on stack
	toLua        ( lua, arg2 );                   		// push 2nd argument on stack
	lua_pcall    ( lua, 2, 1, 0 );                   	// call function taking 2 argsuments and getting one return value
}

template <typename R, typename A1, typename A2, typename A3> 
static inline bool	callLua3 ( lua_State * lua, const char * funcName, R& res, A1 arg1, A2 arg2, A3 arg3 )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );   // push global function f on stack
	toLua        ( lua, arg1 );                   		// push first argument on stack
	toLua        ( lua, arg2 );                   		// push 2nd argument on stack
	toLua        ( lua, arg3 );                   		// push 3rdt argument on stack
	lua_pcall    ( lua, 2, 1, 0 );                   	// call function taking 2 argsuments and getting one return value
	fromLua      ( lua, -1, res );
	lua_pop      ( lua, 1 );
	
	return true;
}

template <typename A1, typename A2, typename A3> 
static inline void	callLua3 ( lua_State * lua, const char * funcName, A1 arg1, A2 arg2, A3 arg3 )
{
	lua_getfield ( lua, LUA_GLOBALSINDEX, funcName );   // push global function f on stack
	toLua        ( lua, arg1 );                   		// push first argument on stack
	toLua        ( lua, arg2 );                   		// push 2nd argument on stack
	toLua        ( lua, arg3 );                   		// push 3rdt argument on stack
	lua_pcall    ( lua, 2, 1, 0 );                   	// call function taking 2 argsuments and getting one return value
}

///////////////////////// template to wrap function for Lua /////////////////////

template <typename R, R (*func)()>
class	WrapFunc0
{
public:
	static int f ( lua_State * lua )
	{
		R	r = func ();
		
		toLua   ( lua, r );
		
		return 1;										// return one vaue
	}
	
	void	registerFunc ( lua_State * lua, const char * name )
	{
		lua_register ( lua, name, f );
	}
};

template <typename R, typename A, R (*func)(A)>
class	WrapFunc1
{
public:
	static int f ( lua_State * lua )
	{
		A	a;
		R	r;
		
		fromLua ( lua, 1, a );
		
		r = func ( a );
		
		toLua   ( lua, r );
		
		return 1;										// return one vaue
	}
	
	void	registerFunc ( lua_State * lua, const char * name )
	{
		lua_register ( lua, name, f );
	}
};

template <typename R, typename A1, typename A2, R (*func)(A1,A2)>
class	WrapFunc2
{
public:
	static int f ( lua_State * lua )
	{
		A1	a1;
		A2	a2;
		R	r;
		
		fromLua ( lua, 1, a1 );
		fromLua ( lua, 2, a2 );
		
		r = func ( a1, a2 );
		
		toLua   ( lua, r );
		
		return 1;										// return one vaue
	}
	
	void	registerFunc ( lua_State * lua, const char * name )
	{
		lua_register ( lua, name, f );
	}
};

template <typename R, typename A1, typename A2, typename A3, R (*func)(A1,A2,A3)>
class	WrapFunc3
{
public:
	static int f ( lua_State * lua )
	{
		A1	a1;
		A2	a2;
		A3	a3;
		R	r;
		
		fromLua ( lua, 1, a1 );
		fromLua ( lua, 2, a2 );
		fromLua ( lua, 3, a3 );
		
		r = func ( a1, a2, a3 );
		
		toLua   ( lua, r );
		
		return 1;										// return one vaue
	}
	
	void	registerFunc ( lua_State * lua, const char * name )
	{
		lua_register ( lua, name, f );
	}
};

#endif

